/*
 * Copyright (c) 2007 Pim Goossens

 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#define _GNU_SOURCE

#include <dlfcn.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <arpa/inet.h>

char *verbose;
int (*real_bind) (int, const struct sockaddr *, socklen_t);

static void port_hack_initialize() __attribute__((constructor));

static void port_hack_initialize()
{
	verbose = getenv("PORT_HACK_VERBOSE");

	if (verbose)
		fputs("port_hack_initialize() called\n", stderr);

	real_bind = dlsym(RTLD_NEXT, "bind");
	if (!real_bind || real_bind == bind)
	{
		fputs("port_hack: error, couldn't find the real bind()!\n", stderr);
		exit(255);
	}
}

int get_port(u_int16_t from_port)
{
	char *ports = getenv("PORT_HACK");
	char from[12];
	int min, max, to_port;
	
	if (!ports)
		return -1;

	do
	{
		if (ports[0] == ',')
		{
			ports = &ports[1];
			if (ports[0] == '\0')
				return -1;
		}

		if (sscanf(ports, "%11[0-9-]:%d", from, &to_port) != 2)
			return -1;

		if (
			sscanf(from, "%d-%d", &min, &max) == 2
			&& min <= from_port && from_port <= max
		   )
			return to_port;

		if (sscanf(from, "%d", &min) == 1 && from_port == min)
			return to_port;
	} while ((ports = strchr(ports, ',')));

	return -1;
}

int bind(int sockfd, const struct sockaddr *my_addr, socklen_t addrlen)
{
	struct sockaddr_in new_addr;

	int from_port, to_port;

	if (!my_addr)
	{
		errno = EFAULT;
		return -1;
	}
	if (my_addr->sa_family != AF_INET)
	{
		if (verbose)
			fputs("port_hack: bind() called for non-INET socket\n", stderr);
		return real_bind(sockfd, my_addr, addrlen);
	}
	new_addr = *((struct sockaddr_in *) my_addr);
	from_port = ntohs(new_addr.sin_port);
	if ((to_port = (int) get_port(from_port)) >= 0)
	{
		if (verbose)
			fprintf(stderr, "port_hack: bind() called for port %d, mapped to %d\n", from_port, to_port);
		new_addr.sin_port = htons((u_int16_t) to_port);
	}
	else if (verbose)
	{
		fprintf(stderr, "bind() called for port %d\n", from_port);
	}

	return real_bind(sockfd, (struct sockaddr *) (&new_addr), addrlen);
}
